/*
 *cryptoslam: A tool for decoding cryptograms
 *
 *Copyright (C) 1997,2000 Brian Enigma
 *                        enigma@netninja.com
 *                        https://sourceforge.net/projects/cryptoslam/
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or EXISTANCE OF KITTEN.  See the GNU General
 *   Public License for more details.
 *
 *   http://www.gnu.org/copyleft/gpl.html
 *
 */

#include "config.h"
#include "cryptogram.h"
#include <string.h>
#include <curses.h>
#include <stdlib.h>
#include <time.h>

Cryptogram::Cryptogram()
{
    cyphertext = NULL;
    plaintext = NULL;
    reset();
}

Cryptogram::~Cryptogram()
{
    if (cyphertext) {free(cyphertext); cyphertext = NULL;}
    if (plaintext)  {free(plaintext);  plaintext  = NULL;}
}

void Cryptogram::reset()
{
    if (cyphertext) {free(cyphertext); cyphertext = NULL;}
    if (plaintext)  {free( plaintext);  plaintext  = NULL;}
    cyphertext = strdup(SAMPLE_TEXT);
    memset(alphabet, ' ', sizeof(alphabet));
    calcPlaintext();
}

void Cryptogram::loadFile(char *filename)
{
    FILE *f;
    char block[4096];
    int size;
    
    if (cyphertext) {free(cyphertext); cyphertext = NULL;}
    if (plaintext)  {free(plaintext);  plaintext  = NULL;}
    cyphertext = strdup(SAMPLE_TEXT);
    memset(alphabet, ' ', sizeof(alphabet));

    f = fopen(filename, "rb");
    if (!f)
    {
        reset();
        return;
    }
    size = fread(block, 1, 4096, f);
    block[4095] = 0;
    block[size] = 0;
    fclose(f);

    if (strncmp(block, FILE_HEADER, strlen(FILE_HEADER) + 1) == 0)
    {
        memcpy(alphabet, block + strlen(FILE_HEADER) + 1, sizeof(alphabet));
        cyphertext = strdup(block + strlen(FILE_HEADER) + 1 + sizeof(alphabet));
    } else {
        cyphertext = strdup(block);
    }
    calcPlaintext();
}

void Cryptogram::saveFileText(char *filename)
{
    FILE *f;
    
    f = fopen(filename, "wb");
    if (!f)
        return;
    fwrite(plaintext, 1, strlen(plaintext), f);
    fwrite("\n\n", 1, 2, f);
    fwrite(cyphertext, 1, strlen(cyphertext), f);
    fclose(f);
}

void Cryptogram::saveFileBinary(char *filename)
{
    FILE *f;
    
    f = fopen(filename, "wb");
    if (!f)
        return;
    // Write the header
    fwrite(FILE_HEADER, 1, strlen(FILE_HEADER) + 1, f);
    // Write the alphabet
    fwrite(alphabet, 1, sizeof(alphabet), f);
    // Write the cyphertext
    fwrite(cyphertext, 1, strlen(cyphertext), f);
    fclose(f);
}

void Cryptogram::set(char cypherLetter, char plainLetter)
{
    // Normalize to uppercase
    cypherLetter = normalize(cypherLetter);
    plainLetter = normalize(plainLetter);
    if (isLetter(cypherLetter))
        alphabet[cypherLetter - 'A'] = plainLetter;
}

void Cryptogram::unset(char cypherLetter)
{
    set(cypherLetter, ' ');
}

void Cryptogram::randomizeKeys()
{
    // Probably a better way of doing this....
    int r1, r2;
    char temp;
    for (int i=0; i<26; i++)
        alphabet[i] = i + 'A';
    srand(time(NULL));
    for (int i=0; i<1000; i++)
    {
        r1 = rand() % 26;
        r2 = rand() % 26;
        temp = alphabet[r1];
        alphabet[r1] = alphabet[r2];
        alphabet[r2] = temp;
    }
}

void Cryptogram::randomizeMessage()
{
    char *tempStr;
    bool matchKey;
    do{
        matchKey = false;
        randomizeKeys();
        for (int i=0; i<26; i++)
        {
            if (alphabet[i]==i+'A')
                matchKey = true;
        }
    }while(matchKey);
    calcPlaintext();
    free(cyphertext);
    cyphertext = plaintext;
    plaintext = strdup(cyphertext);
    calcPlaintext();
    for (int i=0; i<26; i++)
        alphabet[i] = ' ';
}

void Cryptogram::newRandomMessage()
{
    int lines;
    do{
        system("fortune > tmp.tmp");
        loadFile("tmp.tmp");
        lines = 0;
        for (char *c = cyphertext; *c; c++)
            if (*c == '\n')
                lines++;
    }while((lines < 3) || (lines > 8));
    randomizeMessage();
}

void Cryptogram::calcPlaintext()
{
    if (!plaintext)
    {
        plaintext = strdup(cyphertext);
    }
    char *p = plaintext, *c = cyphertext;
    char ch;
    while (*c)
    {
        // Get the character
        ch = *c;
        // Normalize it to uppercase
        ch = normalize(ch);
        // Calc the plaintext
        if (isLetter(ch))
        {
            *p = alphabet[ch - 'A'];
            if (*p == ' ')
                *p = '_';
        } else {
            *p = *c;
        }
        p++;
        c++;
    }
}

void Cryptogram::distribution(char letters[], int counts[])
{
    char *t;
    char letter;
    for (char c = 'A'; c <= 'Z'; c++)
    {
        letters[c-'A'] = c;
        counts[c-'A'] = 0;
    }
    t = cyphertext;
    while (*t)
    {
        letter = normalize(*t);
        if (isLetter(letter))
            counts[letter-'A']++;
        t++;
    }
    // BUBBLESORT!  YEAY!
    int tempInt;
    char tempChar;
    for (int i=0; i < 26; i++)
        for (int j = i; j < 26; j++)
            if (counts[j] > counts[i])
            {
                tempChar = letters[i];
                letters[i] = letters[j];
                letters[j] = tempChar;
                tempInt = counts[i];
                counts[i] = counts[j];
                counts[j] = tempInt;
            }
}

char Cryptogram::normalize(char letter)
{
    if ( (letter >= 'a') && (letter <= 'z') )
        return letter - ('a' - 'A');
    else
        return letter;
}

bool Cryptogram::isLetter(char letter)
{
    bool result = false;
    if ( (letter >= 'A') && (letter <= 'Z') )
        result = true;
    return result;
}

void Cryptogram::paintAlphabet(WINDOW *w)
{
    // Print the alphabet...
    char used[26];
    bool duplicates = false;
    char letter;
    
    for (int i=0; i<26; i++)
        used[i] = 0;

    wcolor_set(w, 2, NULL);
    wmove(w, 0, 0);
    wprintw(w, "Cyphertext: ");
    for (char c='A'; c <= 'Z'; c++)
        wprintw(w, "%c", c);
    wprintw(w, "\nPlaintext:  ");
    for (int i=0; i<26; i++)
    {
        letter = alphabet[i];
        wprintw(w, "%c", letter);
        if (isLetter(letter))
        {
            if (used[letter-'A'] == 1)
                duplicates = true;
            else
                used[letter-'A'] = 1;
        }
    }
    if (duplicates)
        wprintw(w, " <-- Warning: duplicates");
    else
        wprintw(w, "                        ");
    mvwprintw(w, 2, 0, "%s %s by %s", PROGNAME, PROGVER, BYLINE);
}

void Cryptogram::paintMessage(WINDOW *w)
{
    char *here = cyphertext;
    char *newHere;
    char *line;
    // Print the message, interlaced...
    wclear(w);
    wmove(w, 0, 0);
    // Find next \n
    wcolor_set(w, 3, NULL);
    while(*here)
    {
        newHere = strchr(here, '\n');
        if (!newHere)
            newHere = cyphertext + strlen(cyphertext);
        // Print plaintext line
        line = strdup(plaintext + (int)(here - cyphertext));
        for (char *c = line; *c; c++)
            if (*c == '\n') *c = 0;
        wattron(w, A_BOLD);
        wprintw(w, "%s\n", line);
        wattroff(w, A_BOLD);
        free(line);
        // Print cyphertext line
        line = strdup(here);
        for (char *c = line; *c; c++)
            if (*c == '\n') *c = 0;
        wattron(w, A_DIM);
        wprintw(w, "%s\n", line);
        wattroff(w, A_DIM);
        free(line);
        // Move on to next line
        if (newHere)
            here = newHere + 1;
        else
            here = newHere;
    }
}

void Cryptogram::paintStats(WINDOW *w)
{
    // Print the statistics...
    char letters[26];
    int counts[26];
    int row, col;
    distribution(letters, counts);

    wclear(w);
    mvwprintw(w, 1, 0, "Number of times each letter appears in cyphertext:");
    for (int i=0; i < 26; i++)
    {
        if (i < 13)
        {
            row = 3 + i;
            col = 1;
        } else {
            row = 3 + i - 13;
            col = 20;
        }
        wmove(w, row, col);
        wprintw(w, "%c %4d\n", letters[i], counts[i]);
    }
}
